/*  two servo real time clock
 *  ceptimus April 2022
 *
 *  gets time via internet
 *  drives two servos: D5 (minute hand), D6 (hour hand)
 *  servos only need to travel about +/- 68.8 degrees from centre, because there is 34:13 step-up gearing between the servos and the clock hands.
 *  D1 (GPIO5) pull low to drive servos to mid position (for setup)
 *  D2 (GPIO4) pull low to drive servos anti-clockwise to 12 o'clock (for setup)
 *  D7 (GPIO13) pull low to drive servos clockwise to 12 o'clock (for setup)
 */
 
#include <ESP8266WiFi.h>
#include <ESP8266WiFiMulti.h>
#include <WiFiClient.h>
#include <TimeLib.h> //TimeLib library is needed https://github.com/PaulStoffregen/Time

#include <Servo.h>

#include "config.h" // WiFi connection details and time server/timezone set up  ***EDIT THIS FILE PUTTING IN YOUR WIFI DETAILS AND TIMEZONE***

// minute hand servo on GPIO 14 (aka D5)
#define MINUTE_HAND_SERVO 14
// hour hand servo on GPIO 12 (aka D6)
#define HOUR_HAND_SERVO 12

// servo 'driven' time, in milliseconds - servos spend most of the time NOT receiving any pulses, so reducing current draw and wear
#define SERVO_ACTIVE_MS 1000

// next four defines specify servo travel from centre (1500) in microseconds, and direction.  Adjust/reverse these so that the two hands drive exactly one revolution from 12 to 12 clockwise during the start-up test.
// settings below were correct for my Tower Pro SG90 servos. The REVERSE definitions should be the same (both true or both false), unless (strangely) you have mismatched servos: one forward and one reverse rotation.
// if you find that a hand hand actually rotates X minutes instead of the desired 60, then multiply the TRAVEL constant by 60/X; recompile, upload, and try again.
#define MINUTE_TRAVEL 755
#define MINUTE_REVERSE false
#define HOUR_TRAVEL 726
#define HOUR_REVERSE false

// interval for checking time and printing to Serial port (in milliseconds).  Builtin LED is on/off for this time (except when moving a servo) so this is half the blink period
#define UPDATE_INTERVAL 1000

ESP8266WiFiMulti WiFiMulti;

Servo minuteHand;
Servo hourHand;

void setup(void){  
  WiFi.mode(WIFI_STA);
  WiFiMulti.addAP(WIFI_SSID, WIFI_PASSWORD);
    
  // Wait for connection
  while (WiFiMulti.run() != WL_CONNECTED) {
    toggleLED();
    delay(500);
  }

  configTime(0, 0, NTP_TIME_SERVERS);
  setenv("TZ", TIME_ZONE_SETTINGS);
  time_t now;
  
  do {
    toggleLED();
    delay(100);
    time(&now);
  } while (localtime(&now)->tm_year < 120);
  Serial.begin(115200);
  pinMode(5, INPUT_PULLUP); // (D1) short to ground to setup servos mid position (initally set drive gears at mid position, and both clock hands pointing to '6')
  pinMode(4, INPUT_PULLUP); // (D2) short to ground to measure anticlockwise twelve o'clock position
  pinMode(13, INPUT_PULLUP); // (D7) short to ground to measure servos clockwise twelve o'clock position, then edit/upload this sketch till both hands travel exactly one revolution 
  

}

void loop(void) {
  static int prevMinute = -1;  // minute is calculated below, and when it changes, the servo is moved.  -1 forces a change on the first pass
  static int prevHour = -1; // similarly for hour.
    
  static int32_t msToNextUpdate = 0L;
  static uint32_t prevMillis = millis();
  uint32_t ms = millis();
  uint32_t elapsed = ms - prevMillis;
  prevMillis = ms;
  msToNextUpdate -= elapsed;
  if (digitalRead(5) == LOW) { // set servos to mid-position to allow initial positioning of gears at mid position, and hands pointing to '6'
    positionMinuteHand(30);
    positionHourHand(30);
    prevMinute = prevHour = -1; // force setting of both hands following setting phase
  } else if (digitalRead(4) == LOW) { // set servos to 'hands fully anti-clockwise' position, to allow servo travel measurement
    positionMinuteHand(0);
    positionHourHand(0);
    prevMinute = prevHour = -1; // force setting of both hands following setting phase
  } else if (digitalRead(13) == LOW) { // set servos to 'hands fully clockwise' position, to allow servo travel measurement
    positionMinuteHand(60);
    positionHourHand(60);
    prevMinute = prevHour = -1; // force setting of both hands following setting phase
  } else if (msToNextUpdate <= 0L) { // running normally, so get time, print to serial (for debug) toggle LED and possibly move servo(s)
    msToNextUpdate += UPDATE_INTERVAL;
    if (msToNextUpdate <= 0L) { // if still not positive (following one or more of the 'setting' phases)
      msToNextUpdate = UPDATE_INTERVAL;
    }
    
    toggleLED();
    time_t now;
    struct tm * local;
    time(&now);
    local = localtime(&now);
    char bufr[30];
    strftime(bufr, 30, "%a %d %b", local);
    Serial.println(bufr);
  
    strftime(bufr, 30, "%T %Z", local);
  
    char *c;
    for (c = bufr; *c; c++) {
      while (*c == ' ' && *(c + 1) == ' ') { // remove multiple spaces from timezone string, if present
        char *d;
        d = c;
        do {
          d++;
          *(d - 1) = *d;
        } while (*d);
      }
    }
  
    Serial.println(bufr);
    Serial.println();
    
    int minute = local->tm_min;
    if (minute != prevMinute) {
      positionMinuteHand(minute);
      prevMinute = minute;
    }
    int hour = (local->tm_hour) % 12;
    hour = hour * 5 + (minute + 6) / 12; // hour is now not in hour units, but the 'minute position' of the hour hand from 0 to 60
    if (hour != prevHour) {
      positionHourHand(hour);
      prevHour = hour;
    }
  }
}

void toggleLED() {
  static int LEDtoggle = HIGH;
  pinMode(LED_BUILTIN, OUTPUT); // LED is on '16' for diymore ESP8285mcu Strong
  digitalWrite(LED_BUILTIN, LEDtoggle);
  LEDtoggle = LEDtoggle == HIGH ? LOW : HIGH;
}

void positionMinuteHand(int minute) { // minute is from 0 to 60
  minuteHand.attach(MINUTE_HAND_SERVO, 300, 2700);  // limits of 300us and 2700us don't really matter (the servos are driven using .writeMicroseconds()) but set lower and upper limiting values for .writeMicroseconds 
  if (MINUTE_REVERSE) {
    positionServo(minuteHand, minute, 1500 + MINUTE_TRAVEL, 1500 - MINUTE_TRAVEL);
  } else {
    positionServo(minuteHand, minute, 1500 - MINUTE_TRAVEL, 1500 + MINUTE_TRAVEL);    
  }
}

void positionHourHand(int minute) { // works like minute hand, so value is in minutes, not hours
  hourHand.attach(HOUR_HAND_SERVO, 300, 2700);  // limits of 300us and 2700us don't really matter (the servos are driven using .writeMicroseconds()) but set lower and upper limiting values for .writeMicroseconds 
  if (HOUR_REVERSE) {
    positionServo(hourHand, minute, 1500 + HOUR_TRAVEL, 1500 - HOUR_TRAVEL);
  } else {
    positionServo(hourHand, minute, 1500 - HOUR_TRAVEL, 1500 + HOUR_TRAVEL);    
  }
}

void positionServo(Servo servo, int minute, int minuteZero, int minuteSixty) {
  minute = constrain(minute, 0, 60);
  servo.writeMicroseconds(map(minute, 0, 60, minuteZero, minuteSixty));
  delay(SERVO_ACTIVE_MS);
  servo.detach();
}
